""""
Main module that serves as an entry point for Vakt decisions.
Also contains Inquiry class.
"""

import logging

from .util import JsonDumper, PrettyPrint


log = logging.getLogger(__name__)


class Inquiry(JsonDumper, PrettyPrint):
    """Holds all the information about the inquired intent.
    Is responsible to decisions is the inquired intent allowed or not."""

    def __init__(self, resource=None, action=None, subject=None, context=None):
        # explicitly assign empty strings instead of occasional None, (), etc.
        self.resource = resource or ''
        self.action = action or ''
        self.subject = subject or ''
        self.context = context or {}

    @classmethod
    def from_json(cls, data):
        props = cls._parse(data)
        return cls(**props)


class Guard:
    """Executor of policy checks.
       Given a storage and a checker it can decide via `is_allowed` method if a given inquiry allowed or not."""

    def __init__(self, storage, checker):
        self.storage = storage
        self.checker = checker

    def is_allowed(self, inquiry):
        """Is given inquiry intent allowed or not?"""
        try:
            policies = self.storage.find_for_inquiry(inquiry)
            # Storage is not obliged to do the exact policies match. It's up to the storage
            # to decide what policies to return. So we need a more correct programmatically done check.
            answer = self.check_policies_allow(inquiry, policies)
        except Exception:
            log.exception('Unexpected exception occurred while checking Inquiry %s', inquiry)
            answer = False

        if answer:
            log.info('Incoming Inquiry allowed. Data: %s', inquiry)
        else:
            log.info('Incoming Inquiry rejected. Data: %s', inquiry)

        return answer

    def check_policies_allow(self, inquiry, policies):
        """Check if any of a given policy allows a specified inquiry"""
        # If no policies found or None is given -> deny access!
        if not policies:
            return False

        # Filter policies that fit Inquiry by its attributes.
        filtered = [p for p in policies if
                    self.checker.fits(p, 'actions', inquiry.action) and
                    self.checker.fits(p, 'subjects', inquiry.subject) and
                    self.checker.fits(p, 'resources', inquiry.resource) and
                    self.are_rules_satisfied(p, inquiry)]

        # no policies -> deny access!
        # if we have 2 or more similar policies - all of them should have allow effect, otherwise -> deny access!
        return len(filtered) > 0 and all(p.allow_access() for p in filtered)

    @staticmethod
    def are_rules_satisfied(policy, inquiry):
        """Check if rules in the policy are satisfied for a given inquiry's context"""
        for key, rule in policy.rules.items():
            try:
                ctx_rule = inquiry.context[key]
            except KeyError:
                return False
            if not rule.satisfied(ctx_rule, inquiry):
                return False
        return True



"""
All Rules relevant to contexts that work with strings.
It may be subject, object, etc.
"""

import re
import logging

from ..rules.base import Rule


log = logging.getLogger(__name__)


class StringEqualRule(Rule):
    """Rule that is satisfied if the string value equals the specified property of this rule"""

    def __init__(self, val):
        if not isinstance(val, str):
            log.error('%s creation. Initial property should be a string', type(self).__name__)
            raise TypeError('Initial property should be a string')
        self.val = val

    def satisfied(self, what, inquiry=None):
        return isinstance(what, str) and what == self.val


class StringPairsEqualRule(Rule):
    """Rule that is satisfied when given data is an array of pairs and
       those pairs are represented by equal to each other strings"""

    def satisfied(self, what, inquiry=None):
        if not isinstance(what, list):
            return False
        for pair in what:
            if len(pair) != 2:
                return False
            if not isinstance(pair[0], str) and not isinstance(pair[1], str):
                return False
            if pair[0] != pair[1]:
                return False
        return True


class RegexMatchRule(Rule):
    """Rule that is satisfied when given data matches the provided regular expression.
       Note, that you should provide syntactically valid regular-expression string."""

    def __init__(self, pattern):
        try:
            self.regex = re.compile(pattern)
        except Exception as e:
            log.error('%s creation. Failed to compile regexp %s', type(self).__name__, pattern)
            raise TypeError('pattern should be a valid regexp string. Error %s' % e)

    def satisfied(self, what, inquiry=None):
        return bool(self.regex.match(str(what)))



"""
Utility functions and classes for Vakt.
"""

import json
import logging
from abc import abstractmethod


log = logging.getLogger(__name__)


class JsonDumper:
    """
    Mixin for dumping object to JSON
    """

    @classmethod
    @abstractmethod
    def from_json(cls, data):
        """
        Create object from a JSON string
        Returns a new instance of a class
        """
        pass

    def to_json(self, sort=False):
        """
        Get JSON representation of an object
        """
        return json.dumps(self._data(),
                          sort_keys=sort,
                          default=lambda o: o.to_json() if isinstance(o, JsonDumper) else vars(o))

    @classmethod
    def _parse(cls, data):
        """Parse JSON string and return data"""
        try:
            return json.loads(data)
        except ValueError as err:
            log.exception('Error creating %s from json.', cls.__name__)
            raise err

    def _data(self):
        """
        Get the object data. Is useful for overriding in custom classes
        """
        return vars(self)


class PrettyPrint:
    """
    Allows to log objects with all the fields
    """

    def __str__(self):
        return "%s: %s" % (self.__class__, vars(self))



"""
Module for various checkers.
"""

import re
import logging
from functools import lru_cache
from abc import ABCMeta, abstractmethod

from .parser import compile_regex
from .exceptions import InvalidPatternError


log = logging.getLogger(__name__)


class RegexChecker:
    """
    Checker that uses regular expressions.
    E.g. 'Dog', 'Doge', 'Dogs' fit <Dog[se]?>
         'Dogger' doesn't fit <Dog[se]?>
    """

    def __init__(self, cache_size=1024):
        """Set up LRU-cache size for compiled regular expressions."""
        self.compile = lru_cache(maxsize=cache_size)(compile_regex)

    def fits(self, policy, field, what):
        """Does Policy fit the given 'what' value by its 'field' property"""
        where = getattr(policy, field, [])
        for i in where:
            # check if 'where' item is not written in a policy-defined-regex syntax.
            if policy.start_tag not in i and policy.end_tag not in i:
                if i != what:
                    continue       # continue if it's not a string match
                else:
                    return True    # we've found a string match - policy fits by simple string value
            try:
                pattern = self.compile(i, policy.start_tag, policy.end_tag)
            except InvalidPatternError:
                log.exception('Error matching policy, because of failed regex %s compilation', i)
                return False
            if re.match(pattern, what):
                return True
        return False


class StringChecker(metaclass=ABCMeta):
    """
    Checker that uses string equality.
    You have to redefine `compare` method.
    """

    def fits(self, policy, field, what):
        """Does Policy fit the given 'what' value by its 'field' property"""
        where = getattr(policy, field, [])
        for item in where:
            if policy.start_tag == item[0] and policy.end_tag == item[-1]:
                item = item[1:-1]
            if self.compare(what, item):
                return True
        return False

    @abstractmethod
    def compare(self, needle, haystack):
        """Compares two string values. Override it in a subclass"""
        pass


class StringExactChecker(StringChecker):
    """
    Checker that uses exact string equality. Case-sensitive.
    E.g. 'sun' in 'sunny' - False
         'sun' in 'sun' - True
    """

    def compare(self, needle, haystack):
        return needle == haystack


class StringFuzzyChecker(StringChecker):
    """
    Checker that uses fuzzy substring equality. Case-sensitive.
    E.g. 'sun' in 'sunny' - True
         'sun' in 'sun' - True
    """

    def compare(self, needle, haystack):
        return needle in haystack



"""
Exceptions relevant for Vakt workflow.
"""

class InvalidPatternError(Exception):
    """Invalid policy pattern"""
    pass


class PolicyCreationError(Exception):
    """Error during Policy creation occurred."""
    pass


class RuleCreationError(Exception):
    """Error during Rule creation occurred."""
    pass


class PolicyExistsError(Exception):
    """Error when the already existing policy is attempted to be created by Storage"""
    pass



"""
All Rules relevant to networking context
"""

import ipaddress
import logging

from ..rules.base import Rule


log = logging.getLogger(__name__)


class CIDRRule(Rule):
    """Rule that is satisfied when inquiry's IP address is in the provided CIDR"""

    def __init__(self, cidr):
        self.cidr = cidr

    def satisfied(self, what, inquiry=None):
        if not isinstance(what, str):
            return False
        try:
            ip = ipaddress.ip_address(what)
            net = ipaddress.ip_network(self.cidr)
        except ValueError:
            log.exception('Error %s satisfied', type(self).__name__)
            return False
        return ip in net



"""
Base Rule. Should be extended be concrete ones.
"""

import json
import logging
import importlib
from abc import ABCMeta, abstractmethod
from inspect import signature

from ..util import JsonDumper, PrettyPrint
from ..exceptions import RuleCreationError


log = logging.getLogger(__name__)


class Rule(JsonDumper, PrettyPrint, metaclass=ABCMeta):
    """Basic Rule"""

    @abstractmethod
    def satisfied(self, what, inquiry):
        """Is rule satisfied by the inquiry"""
        pass

    @classmethod
    def from_json(cls, data):
        try:
            data = json.loads(data)
        except ValueError as e:
            log.exception('Error creating %s', cls.__name__)
            raise RuleCreationError('Invalid JSON data. JSON error: %s' % e)
        if 'contents' not in data:
            log.exception('Error creating %s. No "contents" key in JSON data found', cls.__name__)
            raise RuleCreationError("No 'contents' key in JSON data found")
        if 'type' not in data:
            log.exception('Error creating %s. No "type" key in JSON data found', cls.__name__)
            raise RuleCreationError("No 'type' key in JSON data found")
        parts = data['type'].split('.')
        module = importlib.import_module(".".join(parts[:-1]))
        klass = getattr(module, parts[-1])

        obj = klass.__new__(klass)
        given_args_len = len(data['contents'])
        expected_args_len = len(signature(klass.__init__).parameters)-1
        if given_args_len != expected_args_len:
            log.exception('Error creating %s. Mismatched number of arguments', cls.__name__)
            raise RuleCreationError(
                'Number of arguments does not match. Given %d. Expected %d' % (given_args_len, expected_args_len))
        for k, v in data['contents'].items():
            setattr(obj, k, v)
        return obj

    def _data(self):
        return {
            'type': '%s.%s' % (self.__class__.__module__, self.__class__.__name__),
            'contents': self.__dict__,
        }



"""
All Rules relevant to Inquiry context
"""

from ..rules.base import Rule


class SubjectEqualRule(Rule):
    """Rule that is satisfied if the string value equals the Inquiry's Subject"""

    def satisfied(self, what, inquiry):
        return isinstance(what, str) and what == inquiry.subject


class ActionEqualRule(Rule):
    """Rule that is satisfied if the string value equals the Inquiry's Action"""

    def satisfied(self, what, inquiry):
        return isinstance(what, str) and what == inquiry.action


class ResourceInRule(Rule):
    """Rule that is satisfied if list contains the Inquiry's Resource"""

    def satisfied(self, what, inquiry):
        return isinstance(what, list) and inquiry.resource in what



from setuptools import setup, find_packages
from os import path

__version__ = '1.0.4'


with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


if __name__ == "__main__":
    setup(
        name='vakt',
        description='Attribute-based access control (ABAC) SDK for Python',
        keywords='ACL RBAC access policy security',
        version=__version__,
        author='Egor Kolotaev',
        author_email='ekolotaev@gmail.com',
        license="Apache 2.0 license",
        url='http://github.com/kolotaev/vakt',
        long_description=long_description,
        long_description_content_type='text/markdown',
        py_modules=['vakt'],
        install_requires=[],
        extras_require={
            'dev': [
                'pytest',
                'pytest-cov',
                'pylint',
            ]
        },
        packages=find_packages(exclude='tests'),
        classifiers=[
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation :: PyPy',
        ],
    )



"""
Functions for parsing and analyzing regex or mixed regex defined rules for Actions and Resources,
"""

import re

from .exceptions import InvalidPatternError


__all__ = ['compile_regex']


def compile_regex(phrase, start_tag, end_tag):
    """Compiles a string denoted by tags to a regular expression"""
    regex_vars, pattern, end = [], '', 0
    indices = get_tag_indices(phrase, start_tag, end_tag)
    for i, idx in enumerate(indices[::2]):
        raw = phrase[end:idx]
        end = indices[i+1]
        part = phrase[idx+1:end-1]
        pattern = pattern + '%s(%s)' % (re.escape(raw), part)
        regex_vars.insert(i//2, re.compile('^%s$' % part))
    raw = phrase[end:]
    return re.compile('^%s%s$' % (pattern, re.escape(raw)))


def get_tag_indices(string, start, end):
    """
    Find and return list of tag indices in the given string.
    Raises exception if tags are not balanced (e.g. <<foo>)
    """
    error_msg = 'Pattern %s has unbalanced braces'
    idx, level = 0, 0
    indices = []
    for i, v in enumerate(string):
        if v == start:
            level = level + 1
            if level == 1:
                idx = i
        elif v == end:
            level = level - 1
            if level == 0:
                indices.append(idx)
                indices.append(i + 1)
            elif level < 0:
                raise InvalidPatternError(error_msg, string)

    if level != 0:
        raise InvalidPatternError(error_msg, string)
    return indices



"""
Namespace for a basic Policy class.
"""

import logging

from .effects import ALLOW_ACCESS, DENY_ACCESS
from .exceptions import PolicyCreationError
from .util import JsonDumper, PrettyPrint
from .rules.base import Rule


log = logging.getLogger(__name__)


class Policy(JsonDumper, PrettyPrint):
    """Represents a policy that regulates access and allowed actions of subjects
    over some resources under a set of rules."""

    def __init__(self, uid, description=None, subjects=(), effect=DENY_ACCESS, resources=(), actions=(), rules=None):
        self.uid = uid
        self.description = description
        self.subjects = subjects
        self.effect = effect or DENY_ACCESS
        self.resources = resources
        self.actions = actions
        rules = rules or {}
        if not isinstance(rules, dict):
            log.exception('Error creating Policy. Rules must be a dictionary')
            raise PolicyCreationError("Error creating Policy. Rules must be a dictionary")
        self.rules = rules

    @classmethod
    def from_json(cls, data):
        props = cls._parse(data)
        if 'uid' not in props:
            log.exception("Error creating policy from json. 'uid' attribute is required")
            raise PolicyCreationError("Error creating policy from json. 'uid' attribute is required")

        rules = {}
        if 'rules' in props:
            for k, clazz in props['rules'].items():
                rules[k] = Rule.from_json(clazz)
        props['rules'] = rules

        return cls(**props)

    def allow_access(self):
        """Does policy imply allow-access?"""
        return self.effect == ALLOW_ACCESS

    @property
    def start_tag(self):
        """Policy expression start tag"""
        return '<'

    @property
    def end_tag(self):
        """Policy expression end tag"""
        return '>'
